"""Functions for performing CARMA analysis on LCs.
"""

import numpy as np
from math import ceil
from scipy.optimize import differential_evolution, minimize
import celerite
from celerite import GP
from eztao.carma.CARMATerm import DRW_term, DHO_term, CARMA_term
from eztao.ts.utils import *

__all__ = [
    "gpSimFull",
    "gpSimRand",
    "gpSimByT",
    "drw_fit",
    "dho_fit",
    "carma_fit",
    "neg_ll",
    "drw_log_param_init",
    "dho_log_param_init",
    "carma_log_param_init",
]


def gpSimFull(carmaTerm, SNR, duration, N, nLC=1):
    """Simulate full CARMA time series.

    Args:
        carmaTerm (object): celerite GP term.
        SNR (float): Signal to noise ratio defined as ratio between
            CARMA amplitude and the mode of the errors (simulated using
            log normal).
        duration (float): The duration of the simulated time series in days.
        N (int): The number of data points.
        nLC (int, optional): Number of light curves to simulate. Defaults to 1.

    Raises:
        Exception: If celerite cannot factorize after 10 trials.

    Returns:
        Arrays: t, y and yerr of the simulated light curves in numpy arrays.
            Note that errors are added to y.
    """

    assert isinstance(
        carmaTerm, celerite.celerite.terms.Term
    ), "carmaTerm must a celerite GP term"

    gp_sim = GP(carmaTerm)

    t = np.linspace(0, duration, N)
    noise = carmaTerm.get_rms_amp() / SNR
    yerr = np.random.lognormal(0, 0.25, N) * noise
    yerr = yerr[np.argsort(np.abs(yerr))]

    # factor and factor_num to track factorization error
    factor = True
    fact_num = 0
    yerr_reg = 1.123e-12

    while factor:
        try:
            gp_sim.compute(t, yerr_reg)
            factor = False
        except Exception:
            # if error, try to re-init t and yerr_reg
            # t = np.linspace(0, duration, N)
            yerr_reg += 1.123e-12

            fact_num += 1
            if fact_num > 10:
                raise Exception(
                    "Celerite cannot factorize the GP"
                    + " covairance matrix, try again!"
                )

    # simulate, assign yerr based on y
    t = np.repeat(t[None, :], nLC, axis=0)
    y = gp_sim.sample(size=nLC)

    # format yerr making it heteroscedastic
    y_rank = y.argsort(axis=1).argsort(axis=1)
    yerr = np.repeat(yerr[None, :], nLC, axis=0)
    yerr = np.array(list(map(lambda x, y: x[y], yerr, y_rank)))
    yerr_sign = np.random.binomial(1, 0.5, yerr.shape)
    yerr_sign[yerr_sign < 1] = -1
    yerr = yerr * yerr_sign

    return t, y + yerr, yerr


def gpSimRand(carmaTerm, SNR, duration, N, nLC=1, season=True, full_N=10_000):
    """Simulate randomly downsampled CARMA time series.

    Args:
        carmaTerm (object): celerite GP term.
        SNR (float): Signal to noise ratio defined as ratio between
            CARMA amplitude and the mode of the errors (simulated using
            log normal).
        duration (float): The duration of the simulated time series in days.
        N (int): The number of data points in the returned light curves.
        nLC (int, optional): Number of light curves to simulate. Defaults to 1.
        season (bool, optional): Whether to simulate seasonal gaps.
            Defaults to True.
        full_N (int, optional): The number of data points the full light curves.
            Defaults to 10_000.

    Returns:
        Arrays: t, y and yerr of the simulated light curves in numpy arrays.
            Note that errors are added to y.
    """
    t, y, yerr = gpSimFull(carmaTerm, SNR, duration, full_N, nLC=nLC)

    # output t & yerr
    tOut = np.empty((nLC, N))
    yOut = np.empty((nLC, N))
    yerrOut = np.empty((nLC, N))

    # downsample
    for i in range(nLC):
        if season:
            mask1 = add_season(t[i])
        else:
            mask1 = np.ones(t[i].shape[0], dtype=np.bool)
        mask2 = downsample_byN(t[i, mask1], N)
        tOut[i, :] = t[i, mask1][mask2]
        yOut[i, :] = y[i, mask1][mask2]
        yerrOut[i, :] = yerr[i, mask1][mask2]

    return tOut, yOut, yerrOut


def gpSimByT(carmaTerm, SNR, t, factor=10, nLC=1):
    """Simulate CARMA time series at the provided timestamps.

    This function uses a 'factor' parameter to determine the sampling rate
    of the initial(full) time series to simulate and downsample from. For
    example, if 'factor' = 10, then the initial time series will be 10 times
    denser than the median sampling rate of the provided timestamps.

    Args:
        carmaTerm (object): celerite GP term.
        SNR (float): Signal to noise ratio defined as ratio between
            CARMA amplitude and the mode of the errors (simulated using
            log normal).
        t (int): Input timestamps.
        factor (int, optional): Paramter to control the ratio in the sampling
            ratebetween the simulated full time series and the desired output.
            Defaults to 10.
        nLC (int, optional): Number of light curves to simulate. Defaults to 1.

    Returns:
        Arrays: t, y and yerr of the simulated light curves in numpy arrays.
            Note that errors are added to y.
    """
    # get number points in full LC based on desired cadence
    duration = ceil(t[-1] - t[0])
    N = factor * ceil(duration / np.median(t[1:] - t[:-1]))

    # simulate full LC
    tFull, yFull, yerrFull = gpSimFull(carmaTerm, SNR, duration, N=N, nLC=nLC)

    # downsample by desired output cadence
    t_expand = np.repeat(t[None, :], nLC, axis=0)
    tOut_idx = np.array(list(map(downsample_byT, tFull, t_expand)))
    tOut = np.array(list(map(lambda x, y: x[y], tFull, tOut_idx)))
    yOut = np.array(list(map(lambda x, y: x[y], yFull, tOut_idx)))
    yerrOut = np.array(list(map(lambda x, y: x[y], yerrFull, tOut_idx)))

    return tOut, yOut, yerrOut


## Below is about Fitting
def neg_ll(params, y, yerr, gp):
    """CARMA neg log likelihood function.

    This method will catch 'overflow/underflow' runtimeWarning and
    return -inf as probablility.

    Args:
        params (object): Array-like, CARMA parameters.
        y (object): Array-like, y values of the time series.
        yerr (object): Array-like, error in y values of the time series.
        gp (object): celerite GP model with the proper kernel.

    Returns:
        float: neg log likelihood.
    """

    # change few runtimewarning action setting
    notify_method = "raise"
    np.seterr(over=notify_method)
    np.seterr(under=notify_method)

    # params = np.array(params)
    dim = params.shape[0]
    run = True
    lap = 0

    while run:
        if lap > 10:
            return -np.inf

        lap += 1
        try:
            gp.set_parameter_vector(params)
            neg_ll = -gp.log_likelihood(y)
            run = False
        except celerite.solver.LinAlgError:
            params += 1e-6 * np.random.randn(dim)
            continue
        except np.linalg.LinAlgError:
            params += 1e-6 * np.random.randn(dim)
            continue
        except FloatingPointError:
            return -np.inf

    return neg_ll


def drw_log_param_init(std):
    """Randomly generate DRW parameters.

    Args:
        std (float): The std of the LC to fit.
    Returns:
        list: The generated DRW parameters in natural log.
    """

    init_tau = np.exp(np.random.uniform(0, 6, 1)[0])
    init_amp = np.random.uniform(0, 4 * std, 1)[0]

    return np.log([init_amp, init_tau])


def dho_log_param_init():
    """Randomly generate DHO parameters.

    Returns:
        list: The generated DHO parameters in natural log.
    """

    log_a1 = np.random.uniform(-10, 1, 1)[0]
    log_a2 = np.random.uniform(-14, -3, 1)[0]
    log_b0 = np.random.uniform(-10, -5, 1)[0]
    log_b1 = np.random.uniform(-10, -5, 1)[0]

    return np.array([log_a1, log_a2, log_b0, log_b1])


def carma_log_param_init(dim):
    """Randomly generate DHO parameters from [-8, 1] in log.

    Args:
        dim (int): For a CARMA(p,q) model, dim=p+q+1.
    Returns:
        list: The generated CAMRA parameters in natural log.
    """

    log_param = np.random.uniform(-8, 2, int(dim))

    return log_param


def _de_opt(y, yerr, best_fit, gp, init_func, debug, bounds):
    """Defferential Evolution optimizer wrapper.

    Args:
        t (object): An array of time stamps in days.
        y (object): An array of y values.
        best_fit (object): An empty array to store best fit parameters.
        gp (object): celerite GP model object.
        init_func ([type]): CARMA parameter initialization function,
            i.e. drw_log_param_init.
        debug (bool, optional): Turn on/off debug mode.
        bounds (list): Initial parameter boundaries for the optimizer.

    Returns:
        object: An array of best-fit parameters
    """

    # dynamic control of fitting flow
    rerun = True
    succeded = False  # ever succeded
    run_ct = 0
    jac_log_rec = 10

    # set bound based on LC std for amp
    while rerun and (run_ct < 5):
        run_ct += 1
        r = differential_evolution(
            neg_ll, bounds=bounds, args=(y, yerr, gp), maxiter=200
        )

        if r.success:
            succeded = True
            best_fit[:] = np.exp(r.x)

            if "jac" not in r.keys():
                rerun = False
            else:
                jac_log = np.log10(np.dot(r.jac, r.jac) + 1e-8)

                # if positive jac, then increase bounds
                if jac_log > 0:
                    bounds = [(x[0] - 1, x[1] + 1) for x in bounds]
                else:
                    rerun = False

                # update best-fit if smaller jac found
                if jac_log < jac_log_rec:
                    jac_log_rec = jac_log
                    best_fit[:] = np.exp(r.x)
        else:
            bounds = [(x[0] - 1, x[1] + 1) for x in bounds]
            gp.set_parameter_vector(init_func())

    # If opitimizer never reached minima, assign nan
    if not succeded:
        best_fit[:] = np.nan

    # Below code is used to visualize if stuck in local minima
    if debug:
        print(r)

    return best_fit


def _min_opt(y, yerr, best_fit, gp, init_func, debug, bounds, method="L-BFGS-B"):
    """A wrapper for scipy.optimize.minimize.

    Args:
        t (object): An array of time stamps in days.
        y (object): An array of y values.
        best_fit (object): An empty array to store best fit parameters.
        gp (object): celerite GP model object.
        init_func ([type]): CARMA parameter initialization function,
            i.e. drw_log_param_init.
        debug (bool, optional): Turn on/off debug mode.
        bounds (list): Initial parameter boundaries for the optimizer.
        method (str): Likelihood optimization method.

    Returns:
        object: An array of best-fit parameters
    """

    # dynamic control of fitting flow
    rerun = True
    succeded = False  # ever succeded
    run_ct = 0
    jac_log_rec = 10

    # set bound based on LC std for amp
    while rerun and (run_ct < 5):
        initial_params = gp.get_parameter_vector()
        run_ct += 1
        r = minimize(
            neg_ll,
            initial_params,
            method=method,
            bounds=bounds,
            args=(y, yerr, gp),
        )
        if r.success:
            succeded = True
            best_fit[:] = np.exp(r.x)

            if "jac" not in r.keys():
                rerun = False
            else:
                jac_log = np.log10(np.dot(r.jac, r.jac) + 1e-8)

                # if positive jac, then increase bounds
                if jac_log > 0:
                    bounds = [(x[0] - 1, x[1] + 1) for x in bounds]
                else:
                    rerun = False

                # update best-fit if smaller jac found
                if jac_log < jac_log_rec:
                    jac_log_rec = jac_log
                    best_fit[:] = np.exp(r.x)
        else:
            bounds = [(x[0] - 1, x[1] + 1) for x in bounds]
            gp.set_parameter_vector(init_func())

    # If opitimizer never reached minima, assign nan
    if not succeded:
        best_fit[:] = np.nan

    # Below code is used to visualize if stuck in local minima
    if debug:
        print(r)

    return best_fit


def drw_fit(t, y, yerr, debug=False, user_bounds=None):
    """Fix time series to DRW model

    Args:
        t (object): An array of time stamps in days.
        y (object): An array of y values.
        yerr (object): An array of the errors in y values.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.
        user_bounds (list, optional): Parameter boundaries for the optimizer.
            Defaults to None.

    Raises:
        Exception: If celerite cannot factorize after 5 trials.

    Returns:
        object: An array of best-fit parameters
    """

    best_fit = np.empty(2)
    std = np.sqrt(np.var(y) - np.var(yerr))

    # init bounds for fitting
    if user_bounds is not None and (len(user_bounds) == 2):
        bounds = user_bounds
    else:
        bounds = [(-4, np.log(4 * std)), (-4, 10)]

    # re-position lc
    t = t - t[0]
    y = y - np.median(y)

    # dynamic control of fitting flow
    compute = True  # handle can't factorize in gp.compute()
    compute_ct = 0

    # initialize parameter and kernel
    kernel = DRW_term(*drw_log_param_init(std))
    gp = GP(kernel, mean=np.median(y))

    # compute can't factorize, try 4 more times
    while compute & (compute_ct < 5):
        compute_ct += 1
        try:
            gp.compute(t, yerr)
            compute = False
        except celerite.solver.LinAlgError:
            if compute_ct > 4:
                raise Exception("celerite can't factorize matrix!")
            gp.set_parameter_vector(drw_log_param_init(std))

    best_fit_return = _de_opt(
        y,
        yerr,
        best_fit,
        gp,
        lambda: drw_log_param_init(std),
        debug,
        bounds,
    )

    return best_fit_return


def dho_fit(t, y, yerr, debug=False, user_bounds=None):
    """Fix time series to DHO model

    Args:
        t (object): An array of time stamps in days.
        y (object): An array of y values.
        yerr (object): An array of the errors in y values.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.
        user_bounds (list, optional): Parameter boundaries for the optimizer.
            Defaults to None.

    Raises:
        Exception: If celerite cannot factorize after 5 trials.

    Returns:
        object: An array of best-fit parameters
    """
    best_fit = np.zeros(4)

    if user_bounds is not None and (len(user_bounds) == 4):
        bounds = user_bounds
    else:
        bounds = [(-10, 7), (-14, 7), (-12, -2), (-11, -2)]

    # re-position lc
    t = t - t[0]
    y = y - np.median(y)

    # dynamic control of fitting flow
    compute = True  # handle can't factorize in gp.compute()
    compute_ct = 0

    # initialize parameter, kernel and GP
    kernel = DHO_term(*dho_log_param_init())
    gp = GP(kernel, mean=np.mean(y))

    # compute can't factorize, try 4 more times
    while compute & (compute_ct < 5):
        compute_ct += 1
        try:
            gp.compute(t, yerr)
            compute = False
        except celerite.solver.LinAlgError:
            if compute_ct > 4:
                raise Exception("celerite can't factorize matrix!")
            gp.set_parameter_vector(dho_log_param_init())

    best_fit_return = _de_opt(
        y,
        yerr,
        best_fit,
        gp,
        lambda: dho_log_param_init(),
        debug,
        bounds,
    )

    return best_fit_return


def carma_fit(t, y, yerr, p, q, de=True, debug=False, user_bounds=None):
    """Fit time series to all CARMA model

    Args:
        t (object): An array of time stamps in days.
        y (object): An array of y values.
        yerr (object): An array of the errors in y values.
        p (int): P order of a CARMA(p, q) model.
        q (int): Q order of a CARMA(p, q) model.
        de (bool, optional): Whether to use differential_evolution as the
            optimizer. Defaults to True.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.
        user_bounds (list, optional): Parameter boundaries for the optimizer.
            Defaults to None.

    Raises:
        Exception: If celerite cannot factorize after 5 trials.

    Returns:
        object: An array of best-fit parameters
    """
    dim = int(p + q + 1)
    best_fit = np.empty(dim)

    # init bounds for fitting
    if user_bounds is not None and (len(user_bounds) == dim):
        bounds = user_bounds
    else:
        bounds = [(-10, 5)] * dim

    # re-position lc
    t = t - t[0]
    y = y - np.median(y)

    # dynamic control of fitting flow
    compute = True  # handle can't factorize in gp.compute()
    compute_ct = 0

    # initialize parameter and kernel
    carma_log_params = carma_log_param_init(dim)
    kernel = CARMA_term(carma_log_params[:p], carma_log_params[p:])
    gp = GP(kernel, mean=np.median(y))

    # compute can't factorize, try 4 more times
    while compute & (compute_ct < 5):
        compute_ct += 1
        try:
            gp.compute(t, yerr)
            compute = False
        except celerite.solver.LinAlgError:
            if compute_ct > 4:
                raise Exception("celerite can't factorize matrix!")
            gp.set_parameter_vector(carma_log_param_init(dim))

    if de:
        best_fit_return = _de_opt(
            y,
            yerr,
            best_fit,
            gp,
            lambda: carma_log_param_init(dim),
            debug,
            bounds,
        )
    else:
        best_fit_return = _min_opt(
            y,
            yerr,
            best_fit,
            gp,
            lambda: carma_log_param_init(dim),
            debug,
            bounds,
        )

    return best_fit_return



import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import celerite
from ..carma.CARMATerm import *

mpl.rc_file("eztao.rc")


def plot_drw_ll(
    t,
    y,
    yerr,
    best_params,
    gp,
    prob_func,
    amp_range=None,
    tau_range=None,
    nLevels=20,
    **kwargs,
):
    """
    Plot DRW likelihood surface

    Args:
        t (array-like): An array of time points
        y (array-like): An array of fluxes at the abvoe time points
        yerr (array-like): An array of photometric errors
        best_params (array-like): Best-fit parameters in [amp, tau]
        gp (object): A DRW celerite GP object
        prob_func (func): Posterior/Likelihood function with 
            args=(params, y, yerr, gp)
        amp_range (array-like, optional): The range of parameters to eval 
            probability. Defaults to None.
        tau_range (array-like, optional): The range of parameters to eval 
            probability. Defaults to None.
        nLevels (int, optional): Contour plot number of levels. Defaults to 20.
    
    Kwargs:
        grid_size (int): The number of points to eval likelihood along a 
            given axis.
        true_params (array-like): The true parameters if the given light curve 
            is simulated.
    """
    best_amp, best_tau = best_params
    grid_size = 50

    if amp_range is None:
        amp_range = [0.5 * best_amp, 1.5 * best_amp]
    if tau_range is None:
        tau_range = [0.5 * best_tau, 1.5 * best_tau]

    # check if custom resoultion
    if "grid_size" in kwargs:
        grid_size = int(kwargs.get("grid_size"))

    taus = np.log(
        np.logspace(np.log10(tau_range[0]), np.log10(tau_range[1]), grid_size)
    )
    amps = np.log(np.linspace(amp_range[0], amp_range[1], grid_size))

    param_grid = np.meshgrid(amps, taus)  # log scale
    reg_param_grid = np.exp(param_grid)  # normal scale

    # flatten to list
    ls_params_combo = list(zip(param_grid[0].flatten(), param_grid[1].flatten()))

    # run in bulk
    gp.compute(t, yerr)
    vec_neg_ll = np.vectorize(prob_func, excluded=[1, 2, 3], signature="(n)->()")
    bulk_lp = np.negative(vec_neg_ll(ls_params_combo, y, yerr, gp))
    lp_reshape = bulk_lp.reshape((grid_size, grid_size))

    # normalize, better contours
    max_ll, min_ll = np.max(bulk_lp), np.min(bulk_lp)
    log_levels = (
        1
        + max_ll
        - np.sort(np.logspace(0, np.log(max_ll - min_ll), nLevels, base=np.e))[::-1]
    )
    norm = mpl.colors.SymLogNorm(
        linthresh=np.abs(max_ll), linscale=1, vmin=min_ll, vmax=max_ll + 1, base=np.e
    )
    fig = plt.figure(figsize=(8, 5), dpi=100)
    _ = plt.contourf(
        reg_param_grid[0],
        reg_param_grid[1],
        lp_reshape,
        log_levels,
        norm=norm,
        extend="both",
    )
    plt.title("Likelihood Surface")
    plt.xlabel("Amp")
    plt.ylabel("Tau")
    plt.scatter(best_params[0], best_params[1], marker="*", s=200, label="Best-fit")

    if "true_params" in kwargs:
        true_params = kwargs.get("true_params")
        plt.scatter(true_params[0], true_params[1], marker="*", s=200, label="True")

    plt.colorbar(_, format="%.0f")
    plt.legend()


def plot_dho_ll(
    t,
    y,
    yerr,
    best_params,
    gp,
    prob_func,
    inner_dim=10,
    outer_dim=4,
    ranges=[(None, None), (None, None), (None, None), (None, None)],
    nLevels=20,
    **kwargs,
):
    """
    Plot DHO likelihood surface

    Args:
        t (array-like): An array of time points
        y (array-like): An array of fluxes at the abvoe time points
        yerr (array-like): An array of photometric errors
        best_params (array-like): Best-fit parameters in [a1, a2, b0, b1]
        gp (object): A DRW celerite GP object
        prob_func (func): Posterior/Likelihood function with 
            args=(params, y, yerr, gp)
        inner_dim (int, optional): The number of points to eval likelihood along 
            a1 and a2. Defaults to 10.
        outer_dim (int, optional): The number of points to eval likelihood along 
            b0 and b1. Defaults to 4.
        ranges (list, optional): Parameters range (log) within which to plot 
            the surface. Defaults to [(None, None), (None, None), (None, None), 
            (None, None)].
        nLevels (int, optional): Contour plot number of levels. Defaults to 20.
   
    Kwargs:
        true_params (array-like): The true parameters in log scale if the 
            given light curve is simulated.
    """
    best_log = np.log(best_params)
    num_param = len(best_log)
    param_range = np.zeros((num_param, 2))

    # check if any ranges provided
    for i in range(num_param):
        if any(map(lambda x: x is None, ranges[i])):
            param_range[i, :] = np.array([best_log[i] - 1, best_log[i] + 1])
        else:
            param_range[i, :] = np.array(ranges[i])

    # create meshgrid to compute ll
    inner_grid_params = [0, 1]
    outer_grid_params = [2, 3]
    grid_ls = []

    for i in range(num_param):
        pm_range = param_range[i]

        # check if param is in inner grid
        if i in inner_grid_params:
            grid_ls.append(np.linspace(pm_range[0], pm_range[1], inner_dim))
        else:
            wide_grid = np.linspace(pm_range[0], pm_range[1], outer_dim + 1)
            delta = wide_grid[1] - wide_grid[0]
            grid_ls.append(wide_grid[1:] - delta / 2)

    param_grid = np.meshgrid(*grid_ls)  # log scale

    # prepare for np.vectorize
    vec_neg_ll = np.vectorize(prob_func, excluded=[1, 2, 3], signature="(n)->()")
    param_combos = zip(
        param_grid[0].flatten(),
        param_grid[1].flatten(),
        param_grid[2].flatten(),
        param_grid[3].flatten(),
    )
    ls_param_combos = list(param_combos)

    # run in bulk
    gp.compute(t, yerr)
    bulk_ll = np.negative(vec_neg_ll(ls_param_combos, y, yerr, gp))

    # back to shape (dim, dim, dim, dim)
    dims = np.empty(num_param, dtype=np.int)
    dims[inner_grid_params] = inner_dim
    dims[outer_grid_params] = outer_dim

    ll_reshape = bulk_ll.reshape(tuple(dims))
    max_ll, min_ll = np.max(bulk_ll), np.min(bulk_ll)
    idx_max = np.unravel_index(
        np.argmax(np.median(ll_reshape, axis=(0, 1)), axis=None), (outer_dim, outer_dim)
    )
    log_levels = (
        1 + max_ll - np.sort(np.logspace(0, np.log10(max_ll - min_ll), nLevels))[::-1]
    )

    # plot
    figsize = (outer_dim * 5, outer_dim * 5)
    fig, axs = plt.subplots(
        outer_dim, outer_dim, figsize=figsize, sharey=True, sharex=True
    )
    images = {}

    b0s = grid_ls[2]
    b1s = grid_ls[3]

    # loop over each grid point on outler layer
    for i in range(outer_dim):
        for j in range(outer_dim):
            _ = axs[i, j].contourf(
                param_grid[0][:, :, i, j],
                param_grid[1][:, :, i, j],
                ll_reshape[:, :, i, j],
                log_levels,
                extend="both",
            )
            images[f"{i}_{j}"] = _
            axs[i, j].scatter(
                best_log[0], best_log[1], marker="*", s=200, label="LL best-fit"
            )
            if "true_params" in kwargs:
                true_params = kwargs.get("true_params")
                axs[i, j].scatter(
                    true_params[0], true_params[1], marker="*", s=200, label="True"
                )

            axs[i, j].set_title(
                f"log(b0):{b0s[i]:.2f};log(b1):{b1s[j]:.2f}", fontsize=18
            )

    # reset the norm
    norm = mpl.colors.SymLogNorm(
        linthresh=np.abs(max_ll), linscale=1, vmin=min_ll, vmax=max_ll + 1, base=10
    )
    for key in images:
        images[key].set_norm(norm)

    fig.colorbar(
        images[f"{idx_max[0]}_{idx_max[1]}"], ax=axs, fraction=0.05, format="%.0f"
    )
    fig.text(0.5, 0.06, "log(a1)", ha="center", fontsize=25)
    fig.text(0.06, 0.5, "log(a2)", va="center", rotation="vertical", fontsize=25)
    axs[idx_max[0], idx_max[1]].legend()


def plot_pred_drw_lc(
    lc_df, best_amp, best_tau, num_data=500, time_col="t", y_col="y", yerr_col="yerr"
):
    """
    Plot GP predicted light curve given best-fit parameters. 

    Args:
        lc_df (dataframe): The dataframe containing the light curve.
        best_amp (float): Best-fit DRW amplitude.
        best_tau (float): Best-fit DRW characteristic timescale.
        num_data (int): The number of points in the predicated LC.
        time_col (str, optional): Time columne name. Defaults to 't'.
        y_col (str, optional): Y axis column name. Defaults to 'y'.
        yerr_col (str, optional): Error column name. Defaults to 'yerr'.
    """

    # get LC data
    lc_df = lc_df.sort_values(by=time_col).reset_index(drop=True)
    t = lc_df[time_col].values - lc_df[time_col].min()
    y = lc_df[y_col].values
    yerr = lc_df[yerr_col].values

    # create GP model using params
    kernel = DRW_term(np.log(best_amp), np.log(best_tau))
    gp = celerite.GP(kernel, mean=np.mean(y))
    gp.compute(t, yerr)

    # generate pred LC
    t_pred = np.linspace(0, np.max(t), num_data)
    return_var = True
    try:
        mu, var = gp.predict(y, t_pred, return_var=return_var)
        std = np.sqrt(var)
    except FloatingPointError as e:
        print(e)
        print("No variance will be returned")
        return_var = False
        mu, var = gp.predict(y, t_pred, return_var=return_var)

    # Plot the data
    fig = plt.figure(figsize=(10, 5))
    color = "#ff7f0e"
    plt.errorbar(t, y, yerr=yerr, fmt=".k", capsize=0, label="Input Data")
    plt.plot(t_pred, mu, color=color, label="Mean Prection")
    if return_var:
        plt.fill_between(
            t_pred, mu + std, mu - std, color=color, alpha=0.3, edgecolor="none"
        )
    plt.ylabel(r"Flux (arb. unit)")
    plt.xlabel(r"Time (day)")
    plt.gca().yaxis.set_major_locator(plt.MaxNLocator(5))
    plt.title("Maximum Likelihood Prediction")
    plt.tight_layout()


def plot_pred_dho_lc(
    lc_df, best_params, num_data=500, time_col="t", y_col="y", yerr_col="yerr"
):
    """
    Plot GP predicted light curve given best-fit parameters. 

    Args:
        lc_df (dataframe): The dataframe containing the light curve.
        best_params (array-like): Best-fit DHO parameters.
        num_data (int): The number of points in the predicated LC.
        time_col (str, optional): Time columne name. Defaults to 't'.
        y_col (str, optional): Y axis column name. Defaults to 'y'.
        yerr_col (str, optional): Error column name. Defaults to 'yerr'.
    """

    # get LC data
    lc_df = lc_df.sort_values(by=time_col).reset_index(drop=True)
    t = lc_df[time_col].values - lc_df[time_col].min()
    y = lc_df[y_col].values
    yerr = lc_df[yerr_col].values

    # create GP model using params
    kernel = DHO_term(*np.log(best_params))
    gp = celerite.GP(kernel, mean=np.mean(y))
    gp.compute(t, yerr)

    # generate pred LC
    t_pred = np.linspace(0, np.max(t), num_data)
    return_var = True
    try:
        mu, var = gp.predict(y, t_pred, return_var=return_var)
        std = np.sqrt(var)
    except FloatingPointError as e:
        print(e)
        print("No variance will be returned")
        return_var = False
        mu, var = gp.predict(y, t_pred, return_var=return_var)

    # Plot the data
    fig = plt.figure(figsize=(10, 5))
    color = "#ff7f0e"
    plt.errorbar(t, y, yerr=yerr, fmt=".k", capsize=0, label="Input Data")
    plt.plot(t_pred, mu, color=color, label="Mean Prection")
    if return_var:
        plt.fill_between(
            t_pred, mu + std, mu - std, color=color, alpha=0.3, edgecolor="none"
        )
    plt.ylabel(r"Flux (arb. unit)")
    plt.xlabel(r"Time (day)")
    plt.gca().yaxis.set_major_locator(plt.MaxNLocator(5))
    plt.title("Maximum Likelihood Prediction")
    plt.tight_layout()


def plot_pred_lc(
    lc_df, best_params, p, q, num_data=500, time_col="t", y_col="y", yerr_col="yerr"
):
    """
    Plot GP predicted light curve given best-fit parameters. 

    Args:
        lc_df (dataframe): The dataframe containing the light curve.
        best_params (array-like): Best-fit CARMA parameters.
        p (int): CARMA p order.
        q (int): CARMA q order.
        num_data (int): The number of points in the predicated LC.
        time_col (str, optional): Time columne name. Defaults to 't'.
        y_col (str, optional): Y axis column name. Defaults to 'y'.
        yerr_col (str, optional): Error column name. Defaults to 'yerr'.
    """

    assert len(best_params) == int(p + q + 1)

    # get LC data
    lc_df = lc_df.sort_values(by=time_col).reset_index(drop=True)
    t = lc_df[time_col].values - lc_df[time_col].min()
    y = lc_df[y_col].values
    yerr = lc_df[yerr_col].values

    # create GP model using params
    kernel = CARMA_term(np.log(best_params[:p]), np.log(best_params[p:]))
    gp = celerite.GP(kernel, mean=np.mean(y))
    compute = True
    compute_ct = 0

    # compute can't factorize, try 4 more times
    while compute & (compute_ct < 5):
        compute_ct += 1
        try:
            gp.compute(t, yerr)
            compute = False
        except celerite.solver.LinAlgError:
            new_params = np.log(best_params) + 1e-6 * np.random.randn(p + q + 1)
            gp.set_parameter_vector(new_params)

    # generate pred LC
    t_pred = np.linspace(0, np.max(t), num_data)
    return_var = True
    try:
        mu, var = gp.predict(y, t_pred, return_var=return_var)
        std = np.sqrt(var)
    except FloatingPointError as e:
        print(e)
        print("No variance will be returned")
        return_var = False
        mu, var = gp.predict(y, t_pred, return_var=return_var)

    # Plot the data
    fig = plt.figure(figsize=(10, 5))
    color = "#ff7f0e"
    plt.errorbar(t, y, yerr=yerr, fmt=".k", capsize=0, label="Input Data")
    plt.plot(t_pred, mu, color=color, label="Mean Prection")
    if return_var:
        plt.fill_between(
            t_pred, mu + std, mu - std, color=color, alpha=0.3, edgecolor="none"
        )
    plt.ylabel(r"Flux (arb. unit)")
    plt.xlabel(r"Time (day)")
    plt.gca().yaxis.set_major_locator(plt.MaxNLocator(5))
    plt.title("Maximum Likelihood Prediction")
    plt.tight_layout()



import numpy as np
from numba import njit, float64, int32, boolean, vectorize, guvectorize
from scipy.spatial import KDTree
from functools import partial

__all__ = ["downsample_byN", "add_season", "downsample_byT"]


def downsample_byN(t, nObs):
    """Utility function to randomly choose N observation from a given light curves

    Args:
        t (array_like): Time stamp of observations in the original light curve.
        N (int): The number of observations in the final light curve.

    Returns:
        An mask of the original length to select data point.
    """
    # random choose index
    idx = np.arange(len(t))
    mask = np.zeros_like(idx, dtype=np.bool_)
    true_idx = np.random.choice(idx, nObs, replace=False)

    # assign chosen index to 1/True
    mask[true_idx] = 1

    return mask


def _get_nearest_idx(tIn, x):
    """Internal function to return nearest neighbor in an array."""
    return (np.abs(tIn - x)).argmin()


def downsample_byT(tIn, tOut):
    """Find the indices of a downsampled light curve given output timestamps.

    Args:
        tIn (object): Numpy array containing the timestamps of the original
            light curve.
        tOut (object): Numpy array containing the timestamps of the ouput
            light curve.

    Returns:
        An numpy array of indices, which can be used to select data points from
            the original light curve. Note that there could be duplicates.
    """
    get_nearest = partial(_get_nearest_idx, tIn)
    return np.array(list(map(get_nearest, tOut)))


def add_season(t, lc_start=0, season_start=90, season_end=270):
    """Utility function to impose seasonal gap in mock light curves

    Args:
        t (array_like): Time stamp of observations in a light curve.
        lc_start (float): Light curve starting day within a year (0 -> 365.25). Default to 0.
        season_start (float): Observing season start day within a year. Default to 90.
        season_end (float): Observing season end day within a year. Default to 270.

    Returns:
        An mask of the original length to select data point.
    """
    t = t - t[0]
    t = t + lc_start

    mask = (np.mod(t, 365.25) > season_start) & (np.mod(t, 365.25) < season_end)

    return mask



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['eztao', 'eztao.carma', 'eztao.ts', 'eztao.viz']

package_data = \
{'': ['*']}

install_requires = \
['celerite>=0.4.0,<0.5.0',
 'matplotlib>=3.3.1,<4.0.0',
 'numba>=0.51.2,<0.52.0',
 'numpy>=1.19.1,<2.0.0',
 'scipy>=1.5.2,<2.0.0']

setup_kwargs = {
    'name': 'eztao',
    'version': '0.1.0',
    'description': 'A toolkit for Active Galactic Nuclei (AGN) time-series analysis.',
    'long_description': '![tests](https://github.com/ywx649999311/EzTao/workflows/tests/badge.svg)\n# EzTao (易道)\n**EzTao** is a toolkit for conducting AGN time-series/variability analysis, mainly utilizing the continuous-time auto-regressive moving average model (CARMA)\n\n## Installation\n```\npip install eztao\n```\n\n#### Dependencies\n>```\n>python = "^3.7"\n>numpy = "^1.19.1"\n>celerite = "^0.4.0"\n>matplotlib = "^3.3.1"\n>scipy = "^1.5.2"\n>numba = "^0.51.2"\n>```\n\n## Development\n`poetry` is used to solve dependencies and to build/publish this package. Below shows how setup the environment for development (assuming you already have `poetry` installed on your machine). \n\n1. Clone this repository, and enter the repository folder.\n2. Create a python virtual environment and activate it. \n    ```\n    python -m venv env\n    source env/bin/activate\n    ```\n3. Install dependencies and **EzTao** in editable mode.\n   ```\n   poetry install\n   ```\n\nNow you should be ready to start adding new features. Be sure to checkout the normal practice regarding how to use `poetry` on its website. When you are ready to push, also make sure the poetry.lock file is checked-in if any dependency has changed. \n',
    'author': 'Weixiang Yu',
    'author_email': 'wy73@drexel.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ywx649999311/EzTao',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import numpy as np
from celerite import terms
from numba import njit, float64, complex128, int32, vectorize

__all__ = ["acf", "DRW_term", "DHO_term", "CARMA_term"]


@njit(complex128[:](complex128[:]))
def _compute_roots(coeffs):
    """Internal jitted function to compute roots"""

    # find roots using np and make roots that are almost real real
    roots = np.roots(coeffs)
    roots[np.abs(roots.imag) < 1e-10] = roots[np.abs(roots.imag) < 1e-10].real

    return roots


@njit(float64[:](float64[:]))
def _compute_exp(params):
    return np.exp(params)


@njit(complex128[:](complex128[:], float64[:], float64[:]))
def acf(arroots, arparam, maparam):
    """Return CARMA ACF coefficients given model parameter in Brockwell et al.
    2001 notation.

    Args:
        arroots (object): AR roots in a numpy array
        arparam (object): AR parameters in a numpy array
        maparam (object): MA parameters in a numpy array

    Returns:
        ACF coefficients in an array, each element correspond to a root
    """
    p = arparam.shape[0]
    q = maparam.shape[0] - 1
    sigma = maparam[0]

    # MA param into Kell's notation
    # arparam = np.array(arparam)
    maparam = np.array([x / sigma for x in maparam])

    # init acf product terms
    num_left = np.zeros(p, dtype=np.complex128)
    num_right = np.zeros(p, dtype=np.complex128)
    denom = -2 * arroots.real + np.zeros_like(arroots) * 1j

    for k in range(q + 1):
        num_left += maparam[k] * np.power(arroots, k)
        num_right += maparam[k] * np.power(np.negative(arroots), k)

    for j in range(1, p):
        root_idx = np.arange(p)
        root_k = arroots[np.roll(root_idx, j)]
        denom *= (root_k - arroots) * (np.conj(root_k) + arroots)

    return sigma ** 2 * num_left * num_right / denom


class DRW_term(terms.Term):
    """Damped Random Walk term.

    Args:
        log_sigma(float): Sigma is the standard deviation of the DRW process.
        log_tau(float): Tau is the characteristic timescale of the DRW process.
    """

    parameter_names = ("log_sigma", "log_tau")

    def get_real_coefficients(self, params):
        log_sigma, log_tau = params

        return (np.exp(2 * log_sigma), 1 / np.exp(log_tau))

    def get_perturb_amp(self):
        """Return the perturbing noise amplitude (b0)."""
        log_sigma, log_tau = self.get_parameter_vector()

        return np.exp((2 * log_sigma - np.log(1 / 2) - log_tau) / 2)

    def get_rms_amp(self):
        """Return the amplitude of CARMA process."""
        log_sigma, log_tau = self.get_parameter_vector()

        return np.exp(log_sigma)


class CARMA_term(terms.Term):
    """General CARMA term with arbitray parameters.

    Args:
        log_arpars (list): The logarithm of AR coefficients.
        log_mapars (list): The logarithm of MA coefficients.
    """

    def __init__(self, log_arpars, log_mapars, *args, **kwargs):

        arpar_temp = "log_a{}"
        mapar_temp = "log_b{}"
        arpar_names = ("log_a1",)
        mapar_names = ("log_b0",)

        self.arpars = _compute_exp(np.array(log_arpars))
        self.mapars = _compute_exp(np.array(log_mapars))
        self.p = len(self.arpars)
        self.q = len(self.mapars) - 1
        # self._roots = _compute_roots(np.append([1 + 0j], self.arpars))

        # combine ar and ma params into one array
        log_pars = np.append(log_arpars, log_mapars)

        # loop over par array to find out how many params
        for i in range(2, self.p + 1):
            arpar_names += (arpar_temp.format(i),)

        for i in range(1, self.q + 1):
            mapar_names += (mapar_temp.format(i),)

        self.parameter_names = arpar_names + mapar_names
        super(CARMA_term, self).__init__(*log_pars, **kwargs)

    def get_real_coefficients(self, params):

        # get roots and acf
        self.arpars = _compute_exp(params[: self.p])
        self.mapars = _compute_exp(params[self.p :])
        self._roots = _compute_roots(np.append([1 + 0j], self.arpars))
        self.acf = acf(self._roots, self.arpars, self.mapars)

        self.mask = np.iscomplex(self._roots)
        acf_real = self.acf[~self.mask]
        roots_real = self._roots[~self.mask]
        num_real = acf_real.shape[0]

        ar = acf_real[:num_real].real
        cr = -roots_real[:num_real].real

        return (ar, cr)

    def get_complex_coefficients(self, params):

        # mask = np.iscomplex(self._roots)
        acf_complex = self.acf[self.mask]
        roots_complex = self._roots[self.mask]

        ac = 2 * acf_complex[::2].real
        bc = 2 * acf_complex[::2].imag
        cc = -roots_complex[::2].real
        dc = -roots_complex[::2].imag

        return (ac, bc, cc, dc)

    def get_rms_amp(self):
        """Return the amplitude of CARMA process."""
        # force to recompute acf
        self.get_all_coefficients()

        return np.sqrt(np.abs(np.sum(self.acf)))


class DHO_term(CARMA_term):
    """Damped Harmonic Oscillator term.

    Args:
        log_a1 (float): The natual logarithm of DHO parameter a1.
        log_a2 (float): The natual logarithm of DHO parameter a2.
        log_b0 (float): The natual logarithm of DHO parameter b0.
        log_b1 (float): The natual logarithm of DHO parameter b1.
    """

    def __init__(self, log_a1, log_a2, log_b0, log_b1, *args, **kwargs):
        """Inits DHO term."""
        super(DHO_term, self).__init__([log_a1, log_a2], [log_b0, log_b1], **kwargs)



"""A set of functions to assist CARMA analysis.

Methods:
    carma_psd: Return a function that computes the PSD of the specified CARMA    
        process at given frequencies. 
    drw_psd: Return a function that computes the PSD of the specified DRW    
        process at given frequencies. 
    gp_psd: Return a function that computes the PSD of the native celerite GP 
        at given frequencies.
    carma_sf: Return a function that computes the structure function of the 
        specified CARMA process at given time lags.
    drw_sf: Return a function that computes the structure function of the 
        specified DRW process at given time lags.
    carma_acf: Return a function that computes the auto-correlation function of the 
        specified CARMA process at given time lags.
    drw_acf: Return a function that computes the auto-correlation function of the 
        specified DRW process at given time lags.
"""
import numpy as np
from .CARMATerm import acf, CARMA_term


def drw_psd(amp, tau):
    """Return a function that computes DRW Power Spectral Density (PSD).

    Args:
        amp (float): DRW amplitude
        tau (float): DRW decorrelation timescale

    Returns:
        A function that takes in frequencies and returns PSD at the 
            given frequencies.
    """

    # convert amp, tau to CARMA parameters
    a0 = 1 / tau
    sigma2 = 2 * amp ** 2 * a0

    def psd(f):
        return sigma2 / (a0 ** 2 + (2 * np.pi * f) ** 2)

    return psd


def carma_psd(arparams, maparams):
    """Return a function that computes Power Spectral Density (PSD).

    Args:
        arparams (object): Numpy array containing AR parameters.
        maparams (object): Numpy array containing MA parameters

    Returns:
        A function that takes in frequencies and returns PSD at the 
            given frequencies.
    """
    arparams = np.insert(arparams, 0, 1)
    maparams = np.array(maparams)
    arparams_rv = arparams[::-1]

    def psd(f):
        # init terms
        num_terms = np.complex(0)
        denom_terms = np.complex(0)

        for i, param in enumerate(maparams):
            num_terms += param * np.power(2 * np.pi * f * (1j), i)

        for k, param in enumerate(arparams_rv):
            denom_terms += param * np.power(2 * np.pi * f * (1j), k)

        num = np.abs(np.power(num_terms, 2))
        denom = np.abs(np.power(denom_terms, 2))

        return num / denom

    return psd


def gp_psd(carmaTerm):
    """Return a function that computes native GP Power Spectral Density (PSD).

    Args:
        carmaTerm (object): A celerite CARMA term.
    Returns:
        A function that takes in frequencies and returns PSD at the 
            given frequencies.
    """

    def psd(f):
        return 2.5 * carmaTerm.get_psd(2 * np.pi * f)

    return psd


def drw_acf(tau):
    """Return a function that computes the DRW autocorrelation function (ACF).

    Args:
        tau (float): DRW decorrelation timescale.

    Returns:
        A function that takes in time lags and returns ACF at the given lags.
    """
    # convert to CARMA parameter
    a0 = 1 / tau

    def acf(lag):
        return np.exp(-a0 * lag)

    return acf


def carma_acf(arparams, maparams):
    """Return a function that computes the model autocorrelation function (ACF).

    Args:
        arparams (list): AR parameters in a list (or array-like)
        maparams (list): MA parameters in a list (or array-like)

    Returns:
        A function that takes in time lags and returns ACF at the given lags.
    """

    autocorr = acf(arparams, maparams)
    roots = np.roots(np.append([1], arparams))
    gpTerm = CARMA_term(np.log(arparams), np.log(maparams))

    def autocorr_func(lag):
        R = 0
        for i, r in enumerate(roots):
            R += autocorr[i] * np.exp(r * lag)

        return R / gpTerm.get_rms_amp() ** 2

    return autocorr_func


def drw_sf(amp, tau):
    """Return a function that computes the structure function (SF) of DRW.

    Args:
        amp (float): DRW variability amplitude
        tau (float): DRW decorrelation timescale.

    Returns:
        A function that takes in time lags and returns SF at the given lags.
    """

    def sf(lag):
        return np.sqrt(amp ** 2 * (1 - drw_acf(tau)(lag)))

    return sf


def carma_sf(arparams, maparams):
    """Return a function that computes the CARMA structure function (SF).

    Args:
        arparams (list): AR parameters in a list (or array-like)
        maparams (list): MA parameters in a list (or array-like)

    Returns:
        A function that takes in time lags and returns SF at the given lags.
    """
    gpTerm = CARMA_term(np.log(arparams), np.log(maparams))
    amp2 = gpTerm.get_rms_amp() ** 2

    def sf(lag):
        return np.sqrt(amp2 * (1 - carma_acf(arparams, maparams)(lag)))

    return sf
